from datetime import date
import requests
import pandas as pd
import pytest

from pylim import lim


def test_lim_query():
    q = 'Show \r\nFB: FB FP: FP when date is after 2019'
    res = lim.query(q)
    assert res is not None
    assert 'FB' in res.columns
    assert 'FP' in res.columns
    assert res.attrs['query'] == q


def test_extended_query():
    q = '''
    LET
    FP = FP(ROLLOVER_DATE = "5 days before expiration day",ROLLOVER_POLICY = "actual prices")
    FP_M2 = FP(ROLLOVER_DATE = "5 days before expiration day",ROLLOVER_POLICY = "2 nearby actual prices")

    SHOW
    FP: FP
    FP_02: FP_M2
    '''
    res = lim.query(q)
    assert res is not None
    assert 'FP' in res.columns
    assert 'FP_02' in res.columns


def test_invalid_query():
    with pytest.raises(requests.HTTPError):
        _ = lim.query('Show 1: Something + Else')


def test_lim_query_as_curve():
    q = '''LET
ATTR curve0 = forward_curve(FB,"Close","LAST","","","days","",0 day ago)
SHOW
Last:curve0 1 month ago
when date is after 2020

        '''
    res = lim.query_as_curve(q)
    assert isinstance(res, pd.DataFrame)


def test_series():
    res = lim.series('FP_2020J', start_date=date(2020, 1, 1))
    assert res['FP_2020J']['2020-01-02'] == 608.5


def test_series2():
    res = lim.series('PA0002779.6.2')
    assert res['PA0002779.6.2']['2020-01-02'] == 479.75


def test_series3():
    res = lim.series('PUMFE03')
    assert res['PUMFE03']['2020-01-01'] == 463.716


def test_series4():
    res = lim.series('PJABA00')
    assert res['PJABA00']['1990-01-02'] == 246.5


def test_series5():
    res = lim.series({'FP_2020J': 'GO', 'FB_2020J': 'Brent'})
    assert res['GO']['2020-01-02'] == 608.5
    assert res['Brent']['2020-01-02'] == 65.56
    assert res.attrs['symbolmap']['GO'] == 'FP_2020J'


def test_curve1():
    res = lim.curve({'FP': 'GO', 'FB': 'Brent'})
    assert 'GO' in res.columns
    assert 'Brent' in res.columns


def test_curve2():
    res = lim.curve({'FP': 'GO', 'FB': 'Brent'})
    assert 'GO' in res.columns
    assert 'Brent' in res.columns


def test_curve3():
    res = lim.curve('FB', curve_dates=(pd.to_datetime('2020-03-17'),))
    assert res['2020/03/17']['2020-05-01'] == 28.73
    assert res['2020/03/17']['2020-08-01'] == 33.25


def test_curve4():
    res = lim.curve('FB', curve_dates=pd.to_datetime('2020-03-17'))
    assert res['2020/03/17']['2020-05-01'] == 28.73
    assert res['2020/03/17']['2020-08-01'] == 33.25


def test_curve_history():
    res = lim.curve('FP', curve_dates=(pd.to_datetime('2020-03-17'), pd.to_datetime('2020-03-18')))
    assert '2020/03/17' in res.columns
    assert '2020/03/18' in res.columns


def test_curve_formula1():
    res = lim.curve_formula(formula='Show 1: FP/7.45-FB')
    assert 'FP' in res.columns
    assert 'FB' in res.columns
    assert '1' in res.columns


def test_curve_formula2():
    cd = (pd.to_datetime('2020-02-02'), pd.to_datetime('2020-04-04'))
    res = lim.curve_formula(formula='Show 1: FP/7.45-FB', curve_dates=cd)
    assert '2020/02/02' in res.columns
    assert '2020/04/04' in res.columns
    assert res['2020/02/02']['2020-05-01'] == 10.929
    assert res['2020/04/04']['2020-08-01'] == 8.50930


def test_symbol_contracts1():
    res = lim.get_symbol_contract_list('FB', monthly_contracts_only=True)
    assert 'FB_1998J' in res
    assert 'FB_2020Z' in res


def test_symbol_contracts2():
    res = lim.get_symbol_contract_list('CL', 'FB', monthly_contracts_only=True)
    assert 'CL_1998J' in res
    assert 'FB_2020Z' in res


def test_futures_contracts1():
    res = lim.contracts('FB', start_year=2020, start_date='2020-01-01')
    assert '2020Z' in res.columns
    assert "%sZ" % date.today().year in res.columns


def test_futures_contracts2():
    res = lim.contracts('FB', months=['Z'], start_date='date is within 5 days')
    assert "%sZ" % date.today().year in res.columns


def test_futures_contracts_formula():
    res = lim.contracts(formula='Show 1: FP/7.45-FB', months=['F'], start_year=2020, start_date='2020-01-01')
    assert "%sF" % date.today().year in res.columns
    assert res['2021F']['2020-01-02'] == pytest.approx(16.95, abs=0.01)


def test_futures_contracts_formula2():
    res = lim.contracts(formula='Show 1: (FP/7.45-FB) * GBPUSD', months=['F'], start_year=2020, start_date='2020-01-01')
    assert "%sF" % date.today().year in res.columns
    assert res['2021F']['2020-01-02'] == pytest.approx(22.2659, abs=0.01)


def test_cont_futures_rollover():
    res = lim.continuous_futures_rollover('FB', months=['M1', 'M12'], after_date=2019)
    assert res['M1'][pd.to_datetime('2020-01-02')] == 66.25
    assert res['M12'][pd.to_datetime('2020-01-02')] == 60.94


def test_structure1():
    res = lim.structure('FB', 1, 2, start_date=pd.to_datetime('2020-01-01'))
    assert res['M1-M2'][pd.to_datetime('2020-01-02')] == pytest.approx(0.689, abs=0.01)


def test_structure2():
    res = lim.structure('FB', 1, 12, start_date=pd.to_datetime('2020-01-01'))
    assert res['M1-M12'][pd.to_datetime('2020-01-02')] == pytest.approx(5.31)


def test_structure3():
    res = lim.structure('Show 1: FP/7.45-FB', 1, 2, start_date=pd.to_datetime('2020-01-01'))
    assert res['M1-M2'][pd.to_datetime('2020-01-02')] == pytest.approx(-0.656, abs=0.01)


def test_structure4():
    res = lim.structure('Show 1: FP/7.45-FB', 1, 12, start_date=pd.to_datetime('2020-01-01'))
    assert res['M1-M12'][pd.to_datetime('2020-01-02')] == pytest.approx(-1.18, abs=0.01)


def test_structure():
    res = lim.candlestick_data('FB')
    assert isinstance(res, pd.DataFrame)


def test_metadata():
    symbols = ('FB', 'PCAAS00', 'PUMFE03', 'PJABA00')
    m = lim.relations(*symbols, show_columns=True, date_range=True)
    assert isinstance(m['FB']['daterange'], pd.DataFrame)
    assert 'FB' in m.columns
    assert 'PCAAS00' in m.columns
    assert 'PUMFE03' in m.columns
    assert 'PJABA00' in m.columns
    assert m['PJABA00']['daterange']['start']['Low'] == pd.to_datetime('1979-09-03')
    assert m['PJABA00']['daterange']['start']['Close'] == pd.to_datetime('2011-01-31')
    assert m['PJABA00']['daterange']['start']['High'] == pd.to_datetime('1979-09-03')


def test_relations1():
    symbols = 'BRENT_DATED', 'NG'
    rels = lim.relations(*symbols)
    assert 'BRENT_DATED' in rels
    assert 'NG' in rels


def test_relations2():
    symbol = 'TopRelation:Futures:Cboe'
    res = lim.relations(symbol, desc=True)
    assert res['Cboe']['description'] == "Chicago Board Options Exchange"


def test_relations3():
    symbol = 'FB,CL'
    res = lim.relations(symbol, show_children=True)
    assert 'FB' in res.loc['children']['FB']['name'].iloc[0]
    assert 'CL' in res.loc['children']['CL']['name'].iloc[0]


def test_find_symbols_in_path1():
    path = 'TopRelation:Futures:Ipe'
    res = lim.find_symbols_in_path(path)
    assert 'WI_Q21' in res
    assert 'FB' in res


def test_find_symbols_in_query():
    q = 'Show 1: FP/7.45-FB'
    r = lim.find_symbols_in_query(q)
    assert r['FP'] == 'FUTURES'
    assert r['FB'] == 'FUTURES'



import re
import typing as t
from datetime import date, datetime, timedelta
from itertools import chain

import dateutil
import pandas as pd

from pylim import limutils

LIM_DATETIME_FORMAT = '%m/%d/%Y'


class LimQueryBuilder:
    let_keyword = 'LET'
    show_keyword = 'SHOW'
    when_keyword = 'WHEN'

    def __init__(self):
        self.lets = []
        self.shows = []
        self.whens = []

    def add_let(self, let: str):
        self.lets.append(let)

    def add_show(self, show: str):
        self.shows.append(show)

    def add_when(self, when: str):
        self.whens.append(when)

    def whens_to_or(self):
        """Convert existing buffer of 'whens' to a single OR statement."""
        or_statement = ' OR\n'.join(self.whens)
        or_statement = f'{{ {or_statement} }}'
        self.whens = [or_statement]

    def __str__(self) -> str:
        return '\n'.join(
            chain(
                (self.let_keyword,),
                self.lets,
                (self.show_keyword,),
                self.shows,
                (self.when_keyword,),
                self.whens,
            )
        )


def is_formula(symbol: str) -> bool:
    lowercase = symbol.lower()
    return lowercase.startswith("show") or lowercase.startswith("let")


def build_when_clause(start_date: t.Union[str, date]) -> str:
    if start_date:
        if isinstance(start_date, str):
            if 'date is within' in start_date.lower():
                return start_date
            else:
                start_date = datetime.strptime(start_date, "%Y-%m-%d")
        previous_date = start_date - timedelta(days=1)
        return f'date is after {previous_date:{LIM_DATETIME_FORMAT}}'
    return ''


def build_series_query(
    symbols: t.Tuple[str, ...],
    metadata: t.Optional[pd.DataFrame] = None,
    start_date: t.Optional[t.Tuple[str, date]] = None,
) -> str:
    symbol_query_parts = ['Show']
    for symbol in symbols:
        qx = f'{symbol}: {symbol}'
        if limutils.check_pra_symbol(symbol):
            use_high_low = False
            if metadata is not None:
                meta = metadata[symbol]['daterange']
                if 'Low' in meta.index and 'High' in meta.index:
                    if 'Close' in meta.index and meta.start.Low < meta.start.Close:
                        use_high_low = True
                    if 'MidPoint' in meta.index and meta.start.Low <= meta.start.MidPoint:
                        use_high_low = True
            if use_high_low:
                qx = f'{symbol}: (High of {symbol} + Low of {symbol})/2'
        symbol_query_parts.append(qx)

    when = build_when_clause(start_date)
    if when:
        symbol_query_parts.append(f'when {when}')
    query = '\n'.join(symbol_query_parts)
    return query


def build_curve_query(
    symbols: t.Tuple[str, ...], curve_date: t.Optional[date] = None, column: str = 'Close', curve_formula_str: t.Optional[str] = None
) -> str:
    """
    Build query for multiple symbols and a single curve date.
    """
    builder = LimQueryBuilder()
    curve_date_filter = 'LAST' if curve_date is None else f'{curve_date:{LIM_DATETIME_FORMAT}}'

    # pick the first futures symbol for use in let clause where non-futures symbols are used
    reference_symbol = [x for x in symbols if symbols[x] == 'FUTURES'][0]

    for symbol in symbols:
        if symbols[symbol] == 'FUTURES':
            builder.add_let(f'ATTR x{symbol} = forward_curve({symbol},"{column}","{curve_date_filter}","","","days","",0 day ago)')
            builder.add_show(f'{symbol}: x{symbol}')
            builder.add_when(f'x{symbol} is DEFINED')
        else:
            builder.add_let(f'ATTR x{symbol} = if x{reference_symbol} is defined then {symbol} ENDIF')
            builder.add_show(f'{symbol}: x{symbol}')
    builder.whens_to_or()

    if curve_formula_str is not None:
        if 'Show' in curve_formula_str or 'show' in curve_formula_str:
            curve_formula_str = curve_formula_str.replace('Show', '').replace('show', '')
        for symbol in symbols:
            curve_formula_str = re.sub(r'\b%s\b' % symbol, r'x%s' % symbol, curve_formula_str)
        builder.add_show(curve_formula_str)

    # When no curve date is specified we get a full history so filter it.
    if curve_date is None:
        last_month = datetime.now() - dateutil.relativedelta.relativedelta(months=1)
        builder.add_when(f'and date is after {last_month:{LIM_DATETIME_FORMAT}}')

    return str(builder)


def build_curve_history_query(
    symbols: t.Tuple[str], curve_dates: t.Tuple[date, ...], column: str = 'Close'
) -> str:
    """
    Build query for a single symbol and multiple curve dates.
    """
    builder = LimQueryBuilder()
    symbol = symbols[0]
    for counter, curve_date in enumerate(curve_dates, start=1):
        builder.add_let(f'ATTR x{counter} = forward_curve({symbol},"{column}","{curve_date:{LIM_DATETIME_FORMAT}}","","","days","",0 day ago)')
        builder.add_show(f'{curve_date:%Y/%m/%d}: x{counter}')
        builder.add_when(f'x{counter} is DEFINED')
    builder.whens_to_or()
    return str(builder)


def build_continuous_futures_rollover_query(
    symbol: str,
    months: t.Tuple[str, ...] = ('M1',),
    rollover_date: str = '5 days before expiration day',
    after_date: t.Optional[int] = None,
) -> str:
    if after_date is None:
        after_date = date.today().year - 1
    builder = LimQueryBuilder()
    builder.add_when(f'Date is after {after_date}')
    for month in months:
        m = int(month[1:])
        if m == 1:
            rollover_policy = 'actual prices'
        else:
            rollover_policy = f'{m} nearby actual prices'
        builder.add_let(f'M{m} = {symbol}(ROLLOVER_DATE = "{rollover_date}",ROLLOVER_POLICY = "{rollover_policy}")')
        builder.add_show(f'M{m}: M{m}')
    return str(builder)


def build_futures_contracts_formula_query(
    formula: str,
    matches: t.Tuple[str, ...],
    contracts: t.Tuple[str, ...],
    start_date: t.Optional[t.Union[str, date]] = None,
) -> str:
    builder = LimQueryBuilder()
    for contract in contracts:
        builder.add_show(f'{contract}: x{contract}')
        t = formula
        for vsym in matches:
            t = re.sub(fr'\b{vsym}\b', f'{vsym}_{contract}', t)
        if 'show' in t.lower():
            t = re.sub(r'\Show 1:', f'ATTR x{contract} = ', t)
        else:
            t = f'ATTR x{contract} = {t}'
        builder.add_let(f'{t}')
    builder.add_when(build_when_clause(start_date))
    return str(builder)


def continuous_convention(clause: str, symbol: str, mx: int) -> str:
    if mx != 1:
        return clause.replace(symbol, f'{symbol}_{mx:02d}')
    return clause


def build_structure_query(
    clause: str, symbols: t.Tuple[str, ...], mx: int, my: int, start_date: t.Optional[date] = None
) -> str:
    cx = clause
    cy = clause
    for match in symbols:
        cx = continuous_convention(cx, match, mx)
        cy = continuous_convention(cy, match, my)

    when = ''
    if start_date is not None:
        when = f' when date is after {start_date:{LIM_DATETIME_FORMAT}}'
    query = f'Show M{mx}-M{my}: ({cx}) - ({cy}){when}'
    return query


def extract_clause(query: str) -> str:
    """
    Given a string like "Show 1: x + y", return "x + y".
    """
    return re.sub(r'Show \w:', '', query)



import logging
import time
from datetime import datetime

import lxml.builder
import pandas as pd
import requests
from lxml import etree

from pylim.core import get_lim_session

sleep_time = 0.5
calltries = 50
upload_headers = {'Content-Type': 'text/xml'}
default_column = 'TopColumn:Price:Close'


def check_upload_status(session: requests.Session, job_id: int):
    response = session.get(f"/rs/upload/jobreport/{job_id}")
    code, msg = '', ''
    root = etree.fromstring(response.content)
    status_el = root.find('status')
    if status_el is not None:
        code_el = status_el.find('code')
        if code_el is not None:
            code = code_el.text
        message_el = status_el.find('message')
        if message_el is not None:
            msg = message_el.text
        if code not in {'200', '201', '300', '302'}:
            logging.warning(f'job id {job_id}: code:{code} msg: {msg}')
    return code, msg


def build_upload_xml(df, dfmeta):
    """
    Converts a dataframe (column headings being the treepath) into an XML that the uploader takes.
    """
    E = lxml.builder.ElementMaker()
    ROOT = E.ExcelData
    ROWS = E.Rows
    xROW = E.Row
    xCOL = E.Col
    xCOLS = E.Cols

    entries = []
    count = 1
    for irow, row in df.iterrows():
        for col, val in row.iteritems():
            if pd.isna(val):
                continue
            tokens = col.split(';')
            treepath = tokens[0]
            column = default_column if len(tokens) == 1 else tokens[1]
            desc = dfmeta.get('description', '')
            if isinstance(irow, pd.Timestamp):
                irow = irow.date()
            erow = xROW(
                xCOLS(
                    xCOL(treepath, num="1"),
                    xCOL(column, num="2"),
                    xCOL(str((irow - datetime(1899, 12, 30).date()).days), num="3"), # excel dateformat
                    xCOL(str(val), num="4"),
                    xCOL(desc, num="5"),
                ),
                num=str(count)
            )
            count = count + 1
            entries.append(erow)

    irow = ROOT()
    xROWS = ROWS()
    for x in entries:
        xROWS.append(x)
    irow.append(xROWS)

    res = lxml.etree.tostring(irow, pretty_print=True)
    return res


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def upload_chunk(session, df, dfmeta, chunk_id: int):
    """
    Upload dataframe to MorningStar.

    :param session: Requests HTTP session to reuse.
    :param df: DataFrame to upload.
    :param dfmeta: DataFrame's metadata.
    """
    params = {
        'username': session.auth[0],
        'parsername': 'DefaultParser',
    }
    res = build_upload_xml(df, dfmeta)
    logging.debug(f'Uploading chunk #{chunk_id} to LIM')
    try:
        response = session.post("/rs/api/upload", data=res, headers=upload_headers, params=params)
    except requests.RequestException:
        logging.error(f'For chunk head: \n{df.head()}')
        logging.error(f'For chunk tail: \n{df.tail()}')
        raise
    root = etree.fromstring(response.content)
    intStatus = root.attrib['intStatus']
    if intStatus == '202':
        job_id = root.attrib['jobID']
        logging.debug(f'Submitted job id: {job_id}')
        for i in range(0, calltries):
            code, msg = check_upload_status(session, job_id)
            if code in {'200', '201', '300', '302'}:
                return msg
            time.sleep(sleep_time)


def upload_series(df, dfmeta, max_chunk_size: int = 1000):
    if not len(df.columns):
        return
    chunk_size = int(round(len(df) / max_chunk_size, 0)) or 1
    with get_lim_session() as session:
        for i, chunk in enumerate(chunks(df, chunk_size), start=1):
            upload_chunk(session, chunk, dfmeta, i)



"""
Logic to interact with /rs/api/datarequests resource.
"""
import logging
import time
import typing as t

import pandas as pd
import requests
from lxml import etree

from pylim.core.session import get_lim_session
from pylim.limutils import build_dataframe

calltries = 50
sleep_seconds = 0.5
endpoint_url = '/rs/api/datarequests'


def query(query_text: str) -> pd.DataFrame:
    with get_lim_session() as session:
        response = session.post(
            endpoint_url,
            data=f'<DataRequest><Query><Text>{query_text}</Text></Query></DataRequest>',
        )
        attempt = 1
        while True:
            root = etree.fromstring(response.content)
            status_code = int(root.attrib["status"])
            if status_code == 100:
                df = build_dataframe(root[0])
                df.attrs['query'] = query_text
                return df
            elif status_code == 130:
                logging.info('No data')
                df = pd.DataFrame()
                df.attrs['query'] = query_text
                return df
            elif status_code == 200:
                job_id = int(root.attrib['id'])
                logging.info(f'Job {job_id} not complete, starting to poll...')
                # Overwrite the response object for the next loop iteration.
                response = session.get(f'{endpoint_url}/{job_id}')
            else:
                status_message = root.attrib["statusMsg"]
                raise requests.HTTPError(
                    f"{status_code} LIM Client Error: {status_message} for url: {response.url}", response=response
                )
            if attempt >= calltries:
                break
            if attempt > 1:
                time.sleep(sleep_seconds)
            attempt += 1
            
    # Loop completed without returning any result.
    raise requests.exceptions.RetryError('Run out of tries')



import logging
from os import getenv
from urllib.parse import urljoin
from urllib.request import getproxies

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

logger = logging.getLogger(__name__)


class BaseUrlSession(requests.Session):
    """
    A Session with a URL that all requests will use as a base.

    .. note::
        Trailing slashes provided in the base URL and the path are important,
        because this class will perform the join using the `urljoin` method. 
    
    Example #1: Calling GET method without a trailing slash in URL.
    .. code-block:: python
        >>> s = BaseUrlSession('https://example.com/resource/')
        >>> r = s.get('sub-resource/' params={'foo': 'bar'})
        >>> print(r.request.url)
        https://example.com/resource/sub-resource/?foo=bar

    Example #2: Calling GET method with a trailing slash in URL.
    .. code-block:: python
        >>> s = sessions.BaseUrlSession('https://example.com/resource/')
        >>> r = s.get('/sub-resource/' params={'foo': 'bar'})
        >>> print(r.request.url)
        https://example.com/sub-resource/?foo=bar
    """

    def __init__(self, base_url: str):
        super().__init__()
        self.base_url = base_url

    def request(self, method: str, url: str, *args, **kwargs):
        """Send the request after generating the complete URL."""
        url = self.create_url(url)
        return super().request(method, url, *args, **kwargs)

    def create_url(self, url: str) -> str:
        """Create the URL based off this partial path."""
        return urljoin(self.base_url, url)


def raise_for_status_and_log(response: requests.Response, *args, **kwargs):
    try:
        response.raise_for_status()
    except requests.RequestException:
        logger.error(f'Response error: Code: {response.status_code} Msg: {response.text}')
        raise


def get_lim_session() -> requests.Session:
    """
    HTTP Session object configured for requesting data from LIM API.
    """
    retry_adapter = HTTPAdapter(
        max_retries=Retry(
            total=3,
            backoff_factor=2,
            allowed_methods=["HEAD", "GET", "OPTIONS"],
            status_forcelist=[429, 500, 502, 503, 504],
        ),
    )
    session = BaseUrlSession(
        getenv("LIMSERVER", "https://rwe.morningstarcommodity.com"),
    )
    session.auth = getenv("LIMUSERNAME", ""), getenv("LIMPASSWORD", "")
    session.headers = {"Content-Type": "application/xml"}
    session.proxies = getproxies()
    session.mount("http://", retry_adapter)
    session.mount("https://", retry_adapter)
    session.hooks["response"] = raise_for_status_and_log
    return session



"""
Low level module for building and executing queries to MorningStar LIM API.

This module handles:
- Connecting and authentication to the LIM server,
- XML request/response schema specific to each LIM endpoint,
- Status codes returned by the API,
- Polling for results in case of long running jobs,
- Retry and exception handling logic.

To use this module, the clients have to make sure that following environment variables are set:
- LIMSERVER - URL to the MorningStar API, defaults to https://rwe.morningstarcommodity.com,
- LIMUSERNAME - login to the MorningStar account,
- LIMPASSWORD - password for the MorningStar account.
"""
from .data import query
from .session import get_lim_session



import typing as t
import datetime
import pandas as pd
import calendar as cal
from commodutil import forwards
from pylim import lim
from pylim import limutils
from pylim import limqueryutils as lqu


def quarterly(symbol: str, quarter: int = 1, start_year=datetime.date.today().year,
              end_year=datetime.date.today().year + 2, start_date=t.Optional[datetime.date]):
    """
    Given a symbol or formula, calculate the quarterly average and return as a series of yearly timeseries
    :param symbol:
    :param quarter:
    :param start_year:
    :param end_year:
    :return:
    """
    cmap = {1: ['F', 'G', 'H'], 2: ['J', 'K', 'M'], 3: ['N', 'Q', 'U'], 4: ['V', 'X', 'Z']}

    if quarter == 0:  # calc Q1,Q2,Q3,Q4
        df = lim.contracts(symbol, start_year=start_year, end_year=end_year, start_date=start_date)
        dfs = []
        for qtr in cmap:  # filter columns for each quarter
            d = df[[x for x in df.columns if x[-1] in cmap[qtr]]]
            d = limutils.convert_lim_contracts_to_datetime(d)
            d = limutils.pivots_contract_by_year(d)
            d = d.rename(columns={x: 'Q%s_%s' % (qtr, x) for x in d.columns})  # eg Q12020
            dfs.append(d)
        return pd.concat(dfs, 1)
    else:
        return calendar(symbol, start_year=start_year, end_year=end_year, months=cmap[quarter], start_date=start_date)


def calendar(symbol, start_year=datetime.date.today().year, end_year=datetime.date.today().year + 2,
             months: t.Optional[t.Tuple[str, ...]] = None,
             start_date: t.Optional[datetime.date] = None):
    """
    Given a symbol or formula, calculate the calendar (yearly) average and return as a series of yearly timeseries
    :param months:
    :param symbol:
    :param quarter:
    :param start_year:
    :param end_year:
    :return:
    """
    df = lim.contracts(symbol, start_year=start_year, end_year=end_year, months=months, start_date=start_date)
    df = limutils.convert_lim_contracts_to_datetime(df)
    return limutils.pivots_contract_by_year(df)


def spread(symbol: str, x: t.Tuple[int, str], y: t.Tuple[int, str], z: t.Optional[t.Tuple[int, str]] = None,
           start_year: datetime.date.today().year = None, end_year: datetime.date.today().year = None,
           start_date: t.Optional[datetime.date] = None) -> pd.DataFrame:
    contracts = lim.contracts(symbol, start_year=start_year, end_year=end_year, months=[x, y, z],
                                 start_date=start_date)
    contracts = limutils.convert_lim_contracts_to_datetime(contracts)

    if z is not None:
        if (isinstance(x, int) or x.isnumeric()) and (isinstance(y, int) or y.isnumeric()) and (isinstance(z, int) or z.isnumeric()):
            return forwards.fly(contracts, int(x), int(y), int(z))

    if (isinstance(x, int) or x.isnumeric()) and (isinstance(y, int) or y.isnumeric()):
        return forwards.time_spreads_monthly(contracts, x, y)

    if isinstance(x, str) and isinstance(y, str):
        x, y = x.upper(), y.upper()
        if x.startswith('Q') and y.startswith('Q'):
            return forwards.time_spreads_quarterly(contracts, x, y)

        if x.startswith('CAL') and y.startswith('CAL'):
            return forwards.cal_spreads(forwards.cal_contracts(contracts))


def fly(symbol: str, x: t.Tuple[int, str], y: t.Tuple[int, str], z: t.Optional[t.Tuple[int, str]],
        start_year: t.Optional[datetime.date] = None,
        end_year=None, start_date: t.Optional[datetime.date] = None):
    return spread(symbol, x, y, z, start_year=start_year, end_year=end_year, start_date=start_date)


def multi_spread(symbol, spreads, start_year: datetime.date.today().year = None,
                 end_year: datetime.date.today().year = None,
                 start_date: t.Optional[datetime.date] = None) -> pd.DataFrame:
    contracts = lim.contracts(symbol, start_year=start_year, end_year=end_year, start_date=start_date)
    contracts = limutils.convert_lim_contracts_to_datetime(contracts)

    dfs = []
    for spread in spreads:
        r = forwards.time_spreads_monthly(contracts, spread[0], spread[1])
        r = r.rename(
            columns={x: '%s%s_%s' % (cal.month_abbr[spread[0]], cal.month_abbr[spread[1]], x) for x in r.columns})
        dfs.append(r)

    res = pd.concat(dfs, 1)
    return res




import pytest

from pylim import limutils


@pytest.mark.parametrize(
    "symbol, result",
    [
        ("FB", False),
        ("AAGXJ00", True),
        ("PGACR00", True),
        ("PA0005643.6.0", True),
    ]
)
def test_pra_symbol(symbol: str, result: bool):
    assert limutils.check_pra_symbol(symbol) == result



import pandas as pd

from pylim import limqueryutils


def test_build_futures_contracts_formula_query():
    f = 'Show 1: FP/7.45-FB'
    m = ['FP', 'FB']
    c = ['2020F', '2020G']
    res = limqueryutils.build_futures_contracts_formula_query(f, m, c)
    assert 'FP_2020F/7.45-FB_2020F' in res
    assert 'FP_2020G/7.45-FB_2020G' in res


def test_build_futures_contracts_formula_query_2():
    f = 'Show 1: FP/7.45-FP_LONGER'
    m = ['FP', 'FP_LONGER']
    c = ['2020F', '2020G']
    res = limqueryutils.build_futures_contracts_formula_query(f, m, c)
    assert 'FP_2020F/7.45-FP_LONGER_2020F' in res
    assert 'FP_2020G/7.45-FP_LONGER_2020G' in res


def test_build_build_curve_query():
    matches = {'FP': 'FUTURES', 'FB': 'FUTURES'}
    formula = 'Show 1: FP/7.45-FB'
    column = 'Close'
    curve_dates = pd.to_datetime('2020-05-01')
    res = limqueryutils.build_curve_query(symbols=matches, curve_date=curve_dates, column=column,
                                        curve_formula_str=formula)

    assert 'ATTR xFP = forward_curve(FP,"Close","05/01/2020","","","days","",0 day ago)' in res
    assert 'ATTR xFB = forward_curve(FB,"Close","05/01/2020","","","days","",0 day ago)' in res
    assert '1: xFP/7.45-xFB' in res


def test_build_build_curve_query_mix_types():
    matches = {'FP': 'FUTURES', 'FB': 'FUTURES', 'GBPUSD' : 'NORMAL'}
    formula = 'Show 1: FP/7.45-FB + USDGBP'
    column = 'Close'
    curve_dates = pd.to_datetime('2020-05-01')
    res = limqueryutils.build_curve_query(symbols=matches, curve_date=curve_dates, column=column,
                                        curve_formula_str=formula)

    assert 'ATTR xFP = forward_curve(FP,"Close","05/01/2020","","","days","",0 day ago)' in res
    assert 'ATTR xFB = forward_curve(FB,"Close","05/01/2020","","","days","",0 day ago)' in res
    assert 'ATTR xGBPUSD = if xFP is defined then GBPUSD ENDIF' in res
    assert '1: xFP/7.45-xFB + USDGBP' in res



import pandas as pd
from datetime import date
from datetime import timedelta

import pytest

from pylim import lim
from pylim import limuploader


def test_upload_series():
    # Arrange.
    spot_price1 = 1.21
    spot_price2 = 4.89
    today = date.today()
    # don't work with weekends so make weekday
    if today.weekday() == 5 or today.weekday() == 6:
        today = today + timedelta(days=-2)

    data = {
        'TopRelation:Test:SPOTPRICE1;TopColumn:Price:Close': spot_price1,
        'TopRelation:Test:SPOTPRICE2': spot_price2,
    }
    upload_df = pd.DataFrame(data, index=[today])
    dfmeta = {'description': 'desc'}
    # Act.
    limuploader.upload_series(upload_df, dfmeta)
    download_df = lim.series(['SPOTPRICE1', 'SPOTPRICE2'])
    # Assert
    assert download_df.loc[f"{today:%Y-%m-%d}"]['SPOTPRICE1'] == pytest.approx(spot_price1)
    assert download_df.loc[f"{today:%Y-%m-%d}"]['SPOTPRICE2'] == pytest.approx(spot_price2)



import re
import typing as t
from datetime import date

import pandas as pd
from lxml import etree

from pylim import limutils
from pylim import limqueryutils
from pylim.core import get_lim_session, query
from pylim.limutils import is_sequence


def series(symbols: t.Union[str, dict, tuple], start_date: t.Optional[t.Union[str, date]] = None) -> pd.DataFrame:
    scall = symbols
    if isinstance(scall, str):
        scall = tuple([scall])
    elif isinstance(scall, dict):
        scall = tuple(scall)

    # Get metadata if we have PRA symbol.
    meta = None
    if any([limutils.check_pra_symbol(x) for x in scall]):
        meta = relations(*scall, show_columns=True, date_range=True)

    q = limqueryutils.build_series_query(scall, meta, start_date=start_date)
    res = query(q)

    if isinstance(symbols, dict):
        res = res.rename(columns=symbols)
        res.attrs['symbolmap'] = {v: k for k, v in symbols.items()}

    return res


def curve(
    symbols: t.Union[str, dict, tuple],
    column: str = 'Close',
    curve_dates: t.Optional[t.Union[date, t.Tuple[date, ...]]] = None,
) -> pd.DataFrame:
    scall = symbols
    if isinstance(scall, str):
        if limqueryutils.is_formula(symbols):
            return curve_formula(symbols, column=column, curve_dates=curve_dates)
        scall = tuple([scall])
    elif isinstance(scall, dict):
        scall = tuple(scall)

    if curve_dates is not None:
        if not is_sequence(curve_dates):
            curve_dates = (curve_dates,)
        q = limqueryutils.build_curve_history_query(scall, curve_dates, column)
    else:
        if is_sequence(curve_dates) and len(curve_dates):
            curve_date = curve_dates[0]
        else:
            curve_date = curve_dates
        q = limqueryutils.build_curve_query({x:'FUTURES' for x in scall}, curve_date, column)
    res = query(q)

    if isinstance(symbols, dict):
        res = res.rename(columns=symbols)
        res.attrs['symbolmap'] = {v: k for k, v in symbols.items()}

    # Reindex dates to start of month.
    if res is not None and len(res) > 0:
        res = res.resample('MS').mean()
        return res


def curve_formula(
    formula: str,
    column: str = 'Close',
    curve_dates: t.Optional[t.Tuple[date, ...]] = None,
    matches: t.Optional[t.Tuple[str, ...]] = None
) -> pd.DataFrame:
    """
    Calculate a forward curve using existing symbols.
    """
    if matches is None:
        matches = find_symbols_in_query(formula)
    if curve_dates is None or (is_sequence(curve_dates) and len(curve_dates) == 1):
        if is_sequence(curve_dates):
            curve_dates = curve_dates[0]
        q = limqueryutils.build_curve_query(symbols=matches, curve_date=curve_dates, column=column, curve_formula_str=formula)
        res = query(q)
        res = res.resample('MS').mean()
        if isinstance(curve_dates, date):
            res = res.rename(columns={'1': curve_dates.strftime("%Y/%m/%d")})
    else:
        dfs, res = [], None
        if not is_sequence(curve_dates):
            curve_dates = [curve_dates]
        for d in curve_dates:
            rx = curve_formula(formula, column=column, curve_dates=(d,), matches=matches)
            if rx is not None:
                rx = rx[[d.strftime("%Y/%m/%d")]]
                dfs.append(rx)
        if len(dfs) > 0:
            res = pd.concat(dfs, 1)
            res = res.dropna(how='all', axis=0)

    return res


def query_as_curve(query_text: str) -> pd.DataFrame:
    """
    Given a LIM query that returns a curve, format the return (drop NaN).

    :param query: A MorningStar LIM query text.
    """
    df = query(query_text)
    df = df.resample('MS').mean()
    df = df.dropna()
    return df


def continuous_futures_rollover(
    symbol: str,
    months: t.Tuple[str, ...] = ('M1',),
    rollover_date: str = '5 days before expiration day',
    after_date: t.Optional[date] = None
) -> pd.DataFrame:
    if after_date is None:
        after_date = date.today().year - 1
    q = limqueryutils.build_continuous_futures_rollover_query(
        symbol, months=months, rollover_date=rollover_date, after_date=after_date
    )
    res = query(q)
    return res


def _contracts(
    formula: str,
    matches: t.Tuple[str, ...],
    contracts_list: t.Tuple[str, ...],
    start_date: t.Optional[date] = None,
) -> pd.DataFrame:
    s = []
    for match in matches:
        r = [x.split('_')[-1] for x in contracts_list if match in x]
        s.append(set(r))

    common_contacts = list(set(s[0].intersection(*s)))

    q = limqueryutils.build_futures_contracts_formula_query(
        formula, matches=matches, contracts=common_contacts, start_date=start_date
    )
    df = query(q)
    return df


def contracts(
    formula: str,
    start_year: t.Optional[int] = None,
    end_year: t.Optional[int] = None,
    months: t.Optional[t.Tuple[str, ...]] = None,
    start_date: t.Optional[date] = None,
) -> pd.DataFrame:
    matched_futures = tuple(
        symbol for symbol, type in find_symbols_in_query(formula).items() if type == "FUTURES"
    )
    contracts_list = get_symbol_contract_list(*matched_futures, monthly_contracts_only=True)
    contracts_list = limutils.filter_contracts(contracts_list, start_year=start_year, end_year=end_year, months=months)
    return _contracts(formula, matches=matched_futures, contracts_list=contracts_list, start_date=start_date)


def structure(symbol: str, mx: int, my: int, start_date: t.Optional[date] = None) -> pd.DataFrame:
    matches = find_symbols_in_query(symbol)
    clause = limqueryutils.extract_clause(symbol)
    q = limqueryutils.build_structure_query(clause, matches, mx, my, start_date)
    res = query(q)
    return res


def candlestick_data(symbol: str, days: int = 90):
    df = query(
        f'Show Close: Close of {symbol} '
        f'High: High of {symbol} '
        f'Low: Low of {symbol} '
        f'Open: Open of {symbol} '
        f'when date is within {days} days'
    )
    return df


def relations(
    *symbols: str,
    show_children: bool = False,
    show_columns: bool = False,
    desc: bool = False,
    date_range: bool = False,
    shorthand: bool = False,
) -> pd.DataFrame:
    """
    Allows you to retrieve schema (metadata) information about MorningStar LIM relations.

    :param symbols: Relation names to retrieve information for.
    :param show_children: Whether the response provides the immediate child relation information.
    :param show_columns: Whether the response provides column and relation-column information.
    :param desc: Whether the response provides the description of each relation node.
    :param date_range: Whether the response provides data-range dates information for each column.
    :param shorthand: Whether the response disables field value population for child relations to
                      accelerate the meta-data fetch process. This flag works only with `show_children=True`.
    """
    symbols_encoded = ','.join(set(symbols))
    url = f'/rs/api/schema/relations/{symbols_encoded}'
    params = {
        'showChildren': str(show_children).lower(),
        'showColumns': str(show_columns).lower(),
        'desc': str(desc).lower(),
        'dateRange': str(date_range).lower(),
        'shorthand': str(shorthand).lower()
    }
    with get_lim_session() as session:
        response = session.get(url, params=params)
    root = etree.fromstring(response.content)
    df = pd.concat([pd.Series(x.values(), index=x.attrib) for x in root], axis=1, sort=False)
    if show_children:
        df = limutils.relinfo_children(df, root)
    if date_range:
        df = limutils.relinfo_daterange(df, root)
    # Make symbol names the header.
    df.columns = df.loc['name']
    return df


def find_symbols_in_path(path: str) -> list:
    """
    Given a path in the LIM tree hierarchy, find all symbols in that path.
    """
    symbols = []
    df = relations(path, show_children=True)

    for col in df.columns:
        children = df[col]['children']
        for _, row in children.iterrows():
            if row.type == 'FUTURES' or row.type == 'NORMAL':
                symbols.append(row['name'])
            elif row.type == 'CATEGORY':
                rec_symbols = find_symbols_in_path(f'{path}:{row["name"]}')
                symbols = symbols + rec_symbols
    return symbols


def get_symbol_contract_list(
    *symbols: str,
    monthly_contracts_only: bool = False,
) -> list:
    """
    Given a symbol pull all futures contracts related to it.
    """
    response = relations(*symbols, show_children=True, shorthand=True).T
    response = response[(response.hasChildren == '1') & (pd.notnull(response.children))].T
    children = response.loc['children']
    children = pd.concat(children.values)
    contracts_list = list(children.name.values)
    if monthly_contracts_only:
        contracts_list = [c for c in contracts_list if re.findall(r'\d\d\d\d\w', c)]
    return contracts_list


def find_symbols_in_query(query: str) -> dict:
    m = re.findall(r'\w[a-zA-Z0-9_.]+', query)
    if 'Show' in m:
        m.remove('Show')
    rel = relations(*m).T
    rel = rel[rel['type'].isin(['FUTURES', 'NORMAL'])]
    if len(rel) > 0:
        return rel['type'].to_dict()
    return {}


if __name__ == '__main__':
    q = 'Show 1: FP/7.45-FB'
    find_symbols_in_query(q)
    # contracts('NYMEX.CU')
    # series('PGABM00')
    fcurves = {
        'FB': 'Brent',
        'FP': 'ULSD',
        'JKM': 'NYMEX.JKM',
        'C3_PROPANE_CIF_ARA_LGE': 'Propane'
    }
    curve(fcurves)


import pandas as pd
import pytest

from pylim import limstrategies


def test_quarterly1():
    res = limstrategies.quarterly('FP', start_date='2020-01-01', start_year=2020)
    assert res[2020]['2020-01-02'] == pytest.approx(614.5, abs=0.1)


def test_quarterly2():
    res = limstrategies.quarterly('Show 1: FP/7.45-FB', start_date='2020-01-01', start_year=2020)
    assert res[2020]['2020-01-02'] == pytest.approx(15.998, abs=0.001)


def test_quarterly_all():
    res = limstrategies.quarterly('Show 1: FP/7.45-FB', quarter=0, start_date=pd.to_datetime('2020-01-01'),
                                    start_year=2020)
    assert res['Q1_2020']['2020-01-02'] == pytest.approx(15.998, abs=0.001)
    assert res['Q2_2020']['2020-01-02'] == pytest.approx(16.064, abs=0.001)
    assert res['Q3_2020']['2020-01-02'] == pytest.approx(16.504, abs=0.001)
    assert res['Q4_2020']['2020-01-02'] == pytest.approx(16.961, abs=0.001)


def test_calendar1():
    res = limstrategies.calendar('FP', start_date='2020-01-01', start_year=2020)
    assert res[2020]['2020-01-02'] == pytest.approx(600.3, abs=0.1)


def test_calendar2():
    res = limstrategies.calendar('Show 1: FP/7.45-FB', start_year=2020)
    assert res[2020]['2020-01-02'] == pytest.approx(16.5, abs=0.1)


def test_spread_standard():
    res = limstrategies.spread('FB', x=1, y=2, start_year=2019, end_year=2020, start_date='2019-01-01')
    assert res[2020]['2019-01-02'] == pytest.approx(-0.09, abs=0.01)


def test_spread_year_increment():
    res = limstrategies.spread('FB', x=12, y=12, start_year=2019, end_year=2021)
    assert res[2020]['2020-01-02'] == pytest.approx(3.1, abs=0.1)


def test_spread_quartely():
    res = limstrategies.spread('FB', x='Q1', y='Q2', start_year=2019, end_year=2021)
    assert res[2020]['2019-01-02'] == pytest.approx(-0.29, abs=0.01)


def test_calendar_spread():
    res = limstrategies.spread('FB', x='CAL19', y='CAL20')
    assert res['CAL 2019-2020']['2019-01-02'] == pytest.approx(-1.094, abs=0.01)


def test_spread_formula_standard():
    res = limstrategies.spread('Show 1: FP/7.45-FB', x=1, y=2, start_year=2019, end_year=2020)
    assert res[2020]['2019-01-02'] == pytest.approx(-0.21, abs=0.01)
    assert res[2019]['2018-01-02'] == pytest.approx(-0.13, abs=0.01)

def test_spread_formula_quarterly():
    res = limstrategies.spread('Show 1: FP/7.45-FB', x='Q1', y='Q2', start_year=2019, end_year=2021)
    assert res[2020]['2019-01-02'] == pytest.approx(-0.38, abs=0.01)
    assert res[2019]['2018-01-02'] == pytest.approx(-0.14, abs=0.01)


def test_spread_formula_calendar():
    res = limstrategies.spread('Show 1: FP/7.45-FB', x='CAL19', y='CAL20')
    assert res['CAL 2019-2020']['2019-01-02'] == pytest.approx(-1.48, abs=0.01)


def test_multi_spread():
    res = limstrategies.multi_spread('Show 1: FP/7.45-FB', spreads=[[6, 6], [12, 12]], start_year=2020,
                                        start_date='2020-01-01')
    assert res['JunJun_2020']['2020-01-02'] == pytest.approx(-0.45, abs=0.01)
    assert res['DecDec_2020']['2020-01-02'] == pytest.approx(0.46, abs=0.01)


def test_fly1():
    res = limstrategies.fly('FB', x='1', y='2', z='3', start_year=2019, end_year=2020, start_date='2019-01-01')
    assert res[2020]['2019-01-02'] == pytest.approx(0.01, abs=0.01)


def test_fly2():
    res = limstrategies.fly('Show 1: FP/7.45-FB', x=1, y=2, z=3, start_year=2019, end_year=2020,
                            start_date='2019-01-01')
    assert res[2020]['2019-01-02'] == pytest.approx(0.023, abs=0.01)
